from typing import Any, Optional
from typing import Sequence

def run_command_line(args: Optional[Sequence[str]] = None) -> None: ...
def set_abort_status(status: Any) -> None: ...
def main() -> None: ...
